/***********************************************************************************
* Adjustable Clock: Plasmoid to show date and time in adjustable format.
* Copyright (C) 2008 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "plasma-adjustableclock.h"

#include <QPainter>
#include <QFontMetrics>
#include <QSizeF>
#include <QTimeLine>
#include <QClipboard>
#include <QHeaderView>

#include <Plasma/Theme>

#include <KGlobalSettings>
#include <KTimeZoneWidget>
#include <KDialog>
#include <KGlobalSettings>
#include <KConfigDialog>
#include <KDatePicker>
#include <KLocale>
#include <KMenu>
#include <KCalendarSystem>

AdjustableClock::AdjustableClock(QObject *parent, const QVariantList &args) : ClockApplet(parent, args)
{
    KGlobal::locale()->insertCatalog("libplasmaclock");

    setHasConfigurationInterface(TRUE);

    resize(150, 75);
}

void AdjustableClock::init()
{
    KConfigGroup configuration = config();

    ClockApplet::init();

    setCurrentTimezone(configuration.readEntry("currentTimezone", localTimezone()));

    m_clipboardFormats.append("%c");
    m_clipboardFormats.append("%F");
    m_clipboardFormats.append("%X");
    m_clipboardFormats.append("%f");
    m_clipboardFormats.append("%x");
    m_clipboardFormats.append("%a %b %e %Y");
    m_clipboardFormats.append("%H:%M:%S");
    m_clipboardFormats.append("%a %b %e %H:%M:%S %Y");
    m_clipboardFormats.append("%Y-%m-%d %H:%M:%S");
    m_clipboardFormats.append("%t");

    m_timeFormat = configuration.readEntry("timeFormat", "<center><big>%H:%M:%S</big><br />\n<small>%d.%m.%Y</small></center>");
    m_toolTipFormat = configuration.readEntry("toolTipFormat", "%Y-%m-%d %H:%M:%S");
    m_fastCopyFormat = configuration.readEntry("fastCopyFormat", "%Y-%m-%d %H:%M:%S");
    m_timeDifference = configuration.readEntry("timeDifference", 0);
    m_clipboardFormats = configuration.readEntry("clipboardFormats", m_clipboardFormats);

    m_clock = new Plasma::Label(this);
    m_clock->nativeWidget()->setTextFormat(Qt::RichText);
    m_clock->nativeWidget()->setTextInteractionFlags(Qt::NoTextInteraction);
    m_clock->nativeWidget()->setMargin(3);
    m_clock->setText(formatDate(QDateTime::currentDateTime(), m_timeFormat));

    dataEngine("time")->connectSource(currentTimezone(), this, 1000, Plasma::NoAlignment);

    m_clipboardMenu = new QAction(SmallIcon("edit-copy"), i18n("C&opy to Clipboard"), this);
    m_clipboardMenu->setMenu(new KMenu());
    m_clipboardMenu->setVisible(!m_clipboardFormats.isEmpty());

    m_actions.append(m_clipboardMenu);

    m_timer = new QTimer(this);
    m_timer->setInterval(1000);

    updateSize();

    connect(m_timer, SIGNAL(timeout()), this, SLOT(updateToolTipContent()));
    connect(m_clipboardMenu, SIGNAL(hovered()), this, SLOT(updateClipboardMenu()));
}

void AdjustableClock::dataUpdated(const QString &source, const Plasma::DataEngine::Data &data)
{
    Q_UNUSED(source);

    m_dateTime = QDateTime(data["Date"].toDate(), data["Time"].toTime());
    m_dateTime = m_dateTime.addSecs(m_timeDifference);

    m_clock->setText(formatDate(m_dateTime, m_timeFormat));
}

void AdjustableClock::constraintsEvent(Plasma::Constraints constraints)
{
    if (constraints & Plasma::SizeConstraint)
    {
        updateSize();
    }
}

void AdjustableClock::createClockConfigurationInterface(KConfigDialog *parent)
{
    KConfigGroup configuration = config();
    QString preview;
    int row;

    QWidget *generalConfiguration = new QWidget();
    m_generalConfigurationUi.setupUi(generalConfiguration);

    QWidget *clipboardActions = new QWidget();
    m_clipboardActionsUi.setupUi(clipboardActions);

    m_generalConfigurationUi.timeFormat->setPlainText(m_timeFormat);
    m_generalConfigurationUi.toolTipFormat->setPlainText(m_toolTipFormat);
    m_generalConfigurationUi.fastCopyFormat->setText(m_fastCopyFormat);
    m_generalConfigurationUi.timeDifference->setValue(m_timeDifference);

    m_clipboardActionsUi.clipboardActionsTable->setAlternatingRowColors(TRUE);
    m_clipboardActionsUi.clipboardActionsTable->setSelectionMode(QAbstractItemView::SingleSelection);
    m_clipboardActionsUi.clipboardActionsTable->setColumnCount(2);
    m_clipboardActionsUi.clipboardActionsTable->setHorizontalHeaderItem(0, new QTableWidgetItem(i18n("Format")));
    m_clipboardActionsUi.clipboardActionsTable->setHorizontalHeaderItem(1, new QTableWidgetItem(i18n("Preview")));
    m_clipboardActionsUi.clipboardActionsTable->setColumnWidth(0, 150);

    QHeaderView *header = m_clipboardActionsUi.clipboardActionsTable->horizontalHeader();
    header->setResizeMode(QHeaderView::Stretch);
    header->resizeSection(1, 150);

    m_clipboardActionsUi.moveUpButton->setIcon(KIcon("arrow-up"));

    m_clipboardActionsUi.moveDownButton->setIcon(KIcon("arrow-down"));

    foreach (const QString &string, m_clipboardFormats)
    {
        row = m_clipboardActionsUi.clipboardActionsTable->rowCount();
        m_clipboardActionsUi.clipboardActionsTable->insertRow(row);
        m_clipboardActionsUi.clipboardActionsTable->setItem(row, 0, new QTableWidgetItem(string));

        preview = formatDate(m_dateTime, string);

        QTableWidgetItem *item = new QTableWidgetItem(preview);
        item->setFlags(0);
        item->setToolTip(preview);

        m_clipboardActionsUi.clipboardActionsTable->setItem(row, 1, item);
    }

    itemSelectionChanged();

    parent->addPage(generalConfiguration, i18n("General"), icon());
    parent->addPage(clipboardActions, i18n("Clipboard actions"), "edit-copy");

    connect(m_generalConfigurationUi.previewButton, SIGNAL(clicked()), this, SLOT(updateClock()));
    connect(m_clipboardActionsUi.addButton, SIGNAL(clicked()), this, SLOT(insertRow()));
    connect(m_clipboardActionsUi.deleteButton, SIGNAL(clicked()), this, SLOT(deleteRow()));
    connect(m_clipboardActionsUi.moveUpButton, SIGNAL(clicked()), this, SLOT(moveRowUp()));
    connect(m_clipboardActionsUi.moveDownButton, SIGNAL(clicked()), this, SLOT(moveRowDown()));
    connect(m_clipboardActionsUi.clipboardActionsTable, SIGNAL(itemSelectionChanged()), this, SLOT(itemSelectionChanged()));
    connect(m_clipboardActionsUi.clipboardActionsTable, SIGNAL(cellChanged(int, int)), this, SLOT(updateRow(int, int)));
}

void AdjustableClock::clockConfigAccepted()
{
    KConfigGroup configuration = config();

    m_timeFormat = m_generalConfigurationUi.timeFormat->toPlainText();
    m_toolTipFormat = m_generalConfigurationUi.toolTipFormat->toPlainText();
    m_fastCopyFormat = m_generalConfigurationUi.fastCopyFormat->text();
    m_timeDifference = m_generalConfigurationUi.timeDifference->value();

    configuration.writeEntry("timeFormat", m_timeFormat);
    configuration.writeEntry("toolTipFormat", m_toolTipFormat);
    configuration.writeEntry("fastCopyFormat", m_fastCopyFormat);
    configuration.writeEntry("timeDifference", m_timeDifference);

    m_clipboardFormats.clear();

    for (int i = 0; i < m_clipboardActionsUi.clipboardActionsTable->rowCount(); ++i)
    {
        m_clipboardFormats.append(m_clipboardActionsUi.clipboardActionsTable->item(i, 0)->text());
    }

    configuration.writeEntry("clipboardFormats", m_clipboardFormats);

    m_clipboardMenu->setVisible(!m_clipboardFormats.isEmpty());

    updateClock();

    emit configNeedsSaving();
}

void AdjustableClock::itemSelectionChanged()
{
    QList<QTableWidgetItem *> selectedItems = m_clipboardActionsUi.clipboardActionsTable->selectedItems();

    m_clipboardActionsUi.moveUpButton->setEnabled(!selectedItems.isEmpty() && m_clipboardActionsUi.clipboardActionsTable->row(selectedItems.first()) != 0);
    m_clipboardActionsUi.moveDownButton->setEnabled(!selectedItems.isEmpty() && m_clipboardActionsUi.clipboardActionsTable->row(selectedItems.last()) != (m_clipboardActionsUi.clipboardActionsTable->rowCount() - 1));
    m_clipboardActionsUi.deleteButton->setEnabled(!selectedItems.isEmpty());
}

void AdjustableClock::insertRow()
{
    int row = (m_clipboardActionsUi.clipboardActionsTable->rowCount()?m_clipboardActionsUi.clipboardActionsTable->currentRow():0);

    m_clipboardActionsUi.clipboardActionsTable->insertRow(row);

    m_clipboardActionsUi.clipboardActionsTable->setItem(row, 0, new QTableWidgetItem(QString("")));

    QTableWidgetItem *item = new QTableWidgetItem(QString(""));
    item->setFlags(0);

    m_clipboardActionsUi.clipboardActionsTable->setItem(row, 1, item);

    m_clipboardActionsUi.clipboardActionsTable->setCurrentCell(row, 0);
}

void AdjustableClock::deleteRow()
{
    m_clipboardActionsUi.clipboardActionsTable->removeRow(m_clipboardActionsUi.clipboardActionsTable->row(m_clipboardActionsUi.clipboardActionsTable->selectedItems().at(0)));
}

void AdjustableClock::moveRow(bool up)
{
    int sourceRow = m_clipboardActionsUi.clipboardActionsTable->row(m_clipboardActionsUi.clipboardActionsTable->selectedItems().at(0));
    int destinationRow = (up?(sourceRow - 1):(sourceRow + 1));

    QList<QTableWidgetItem*> sourceItems;
    QList<QTableWidgetItem*> destinationItems;

    for (int i = 0; i < 2; ++i)
    {
        sourceItems.append(m_clipboardActionsUi.clipboardActionsTable->takeItem(sourceRow, i));
        destinationItems.append(m_clipboardActionsUi.clipboardActionsTable->takeItem(destinationRow, i));
    }

    for (int i = 0; i < 2; ++i)
    {
        m_clipboardActionsUi.clipboardActionsTable->setItem(sourceRow, i, destinationItems.at(i));
        m_clipboardActionsUi.clipboardActionsTable->setItem(destinationRow, i, sourceItems.at(i));
    }

    m_clipboardActionsUi.clipboardActionsTable->setCurrentCell(destinationRow, 0);
}

void AdjustableClock::moveRowUp()
{
    moveRow(TRUE);
}

void AdjustableClock::moveRowDown()
{
    moveRow(FALSE);
}

void AdjustableClock::updateRow(int row, int column)
{
    Q_UNUSED(column);

    if (!m_clipboardActionsUi.clipboardActionsTable->item(row, 1))
    {
        return;
    }

    QString preview = formatDate(m_dateTime, m_clipboardActionsUi.clipboardActionsTable->item(row, 0)->text());

    m_clipboardActionsUi.clipboardActionsTable->item(row, 1)->setText(preview);
    m_clipboardActionsUi.clipboardActionsTable->item(row, 1)->setToolTip(preview);
}

void AdjustableClock::updateClock()
{
    m_timeFormat = m_generalConfigurationUi.timeFormat->toPlainText();

    m_clock->setText(formatDate(m_dateTime, m_timeFormat));

    updateSize();
}

void AdjustableClock::updateClipboardMenu()
{
    KMenu *menu = new KMenu();

    foreach (const QString &string, m_clipboardFormats)
    {
        menu->addAction(formatDate(m_dateTime, string));
    }

    connect(menu, SIGNAL(triggered(QAction*)), this, SLOT(copyToClipboard(QAction*)));

    m_clipboardMenu->setMenu(menu);
}

void AdjustableClock::copyToClipboard(QAction* action)
{
    QApplication::clipboard()->setText(action->text());
}

void AdjustableClock::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    if (event->buttons() == Qt::MidButton)
    {
        QApplication::clipboard()->setText(formatDate(m_dateTime, m_fastCopyFormat));

        event->setAccepted(TRUE);

        return;
    }

    ClockApplet::mousePressEvent(event);
}

void AdjustableClock::updateSize()
{
    m_clock->nativeWidget()->adjustSize();

    setPreferredSize(m_clock->nativeWidget()->sizeHint());

    if (formFactor() != Plasma::Horizontal && formFactor() != Plasma::Vertical)
    {
        resize(m_clock->nativeWidget()->sizeHint());
    }
}

void AdjustableClock::toolTipAboutToShow()
{
    updateToolTipContent();

    m_timer->start();
}

void AdjustableClock::toolTipHidden()
{
    Plasma::ToolTipManager::self()->clearContent(this);

    m_timer->stop();
}

void AdjustableClock::updateToolTipContent()
{
    Plasma::ToolTipContent toolTipData;

    toolTipData.setImage(KIcon("chronometer").pixmap(IconSize(KIconLoader::Desktop)));
    toolTipData.setSubText(formatDate(m_dateTime, m_toolTipFormat));
    toolTipData.setAutohide(FALSE);

    Plasma::ToolTipManager::self()->setContent(this, toolTipData);
}

QString AdjustableClock::formatDate(const QDateTime DateTime, QString format)
{
    if (format.isEmpty())
    {
        return "";
    }

    QString string;
    int length = (format.length() - 1);

    for (int index = 0; index < length; ++index)
    {
        if (format.at(index) == '%' && format.at(index + 1) != '%')
        {
            index++;

            switch (format.at(index).unicode())
            {
                case 'a': // weekday, short form
                    string.append(KGlobal::locale()->calendar()->weekDayName(DateTime.date(), KCalendarSystem::ShortDayName));
                break;
                case 'A': // weekday, long form
                    string.append(KGlobal::locale()->calendar()->weekDayName(DateTime.date(), KCalendarSystem::LongDayName));
                break;
                case 'b': // month, short form
                    string.append(KGlobal::locale()->calendar()->monthName(KGlobal::locale()->calendar()->month(DateTime.date()), KGlobal::locale()->calendar()->year(DateTime.date()), ((KGlobal::locale()->nounDeclension() && KGlobal::locale()->dateMonthNamePossessive())?KCalendarSystem::ShortNamePossessive:KCalendarSystem::ShortName)));
                break;
                case 'B': // month, long form
                    string.append(KGlobal::locale()->calendar()->monthName(KGlobal::locale()->calendar()->month(DateTime.date()), KGlobal::locale()->calendar()->year(DateTime.date()), ((KGlobal::locale()->nounDeclension() && KGlobal::locale()->dateMonthNamePossessive())?KCalendarSystem::LongNamePossessive:KCalendarSystem::LongName)));
                break;
                case 'c': // date and time representation, short
                    string.append(KGlobal::locale()->formatDateTime(m_dateTime, KLocale::LongDate));
                break;
                case 'C': // date and time representation, long
                    string.append(KGlobal::locale()->formatDateTime(m_dateTime, KLocale::ShortDate));
                break;
                case 'd': // day of the month, two digits
                    string.append(KGlobal::locale()->calendar()->dayString(DateTime.date(), KCalendarSystem::LongFormat));
                break;
                case 'e': // day of the month, one digit
                    string.append(KGlobal::locale()->calendar()->dayString(DateTime.date(), KCalendarSystem::ShortFormat));
                break;
                case 'f': // date representation, short
                    string.append(KGlobal::locale()->formatDate(m_dateTime.date(), KLocale::ShortDate));
                break;
                case 'F': // time representation, short
                    string.append(KGlobal::locale()->formatTime(m_dateTime.time(), FALSE));
                break;
                case 'H': // hour, 24h format
                    if (m_dateTime.time().hour() < 10)
                    {
                        string.append('0');
                    }

                    string.append(QString::number(m_dateTime.time().hour()));
                break;
                case 'I': // hour, 12h format
                    if ((((m_dateTime.time().hour() + 11) % 12) + 1) < 10)
                    {
                        string.append('0');
                    }

                    string.append(QString::number(((m_dateTime.time().hour() + 11) % 12) + 1));
                break;
                case 'j': // day of the year
                    string.append(QString::number(KGlobal::locale()->calendar()->dayOfYear(DateTime.date())));
                break;
                case 'k': // hour, 24h format, one digit
                    string.append(QString::number(m_dateTime.time().hour()));
                break;
                case 'l': // hour, 12h format, one digit
                    string.append(QString::number(((m_dateTime.time().hour() + 11) % 12) + 1));
                break;
                case 'm': // month, two digits
                    string.append(KGlobal::locale()->calendar()->monthString(DateTime.date(), KCalendarSystem::LongFormat));
                break;
                case 'M': // minute, two digits
                    if (m_dateTime.time().minute() < 10)
                    {
                        string.append('0');
                    }

                    string.append(QString::number(m_dateTime.time().minute()));
                break;
                case 'n': // month, one digit
                    string.append(KGlobal::locale()->calendar()->monthString(DateTime.date(), KCalendarSystem::ShortFormat));
                break;
                case 'p': // pm or am
                    string.append((m_dateTime.time().hour() >= 12)?i18n("pm"):i18n("am"));
                break;
                case 'S': // second, two digits
                    if (m_dateTime.time().second() < 10)
                    {
                        string.append('0');
                    }

                    string.append(QString::number(m_dateTime.time().second()));
                break;
                case 't': // UNIX timestamp
                    string.append(QString::number(m_dateTime.toTime_t()));
                break;
                case 'w': // day of week
                    string.append(KGlobal::locale()->calendar()->dayOfWeek(DateTime.date()));
                break;
                case 'W': // week number
                case 'U':
                    string.append(QString::number(KGlobal::locale()->calendar()->weekNumber(DateTime.date())));
                break;
                case 'x': // date representation, long
                    string.append(KGlobal::locale()->formatDate(m_dateTime.date(), KLocale::LongDate));
                break;
                case 'X': // time representation, long
                    string.append(KGlobal::locale()->formatTime(m_dateTime.time(), TRUE));
                break;
                case 'Y': // year, four digits
                    string.append(KGlobal::locale()->calendar()->yearString(DateTime.date(), KCalendarSystem::LongFormat));
                break;
                case 'y': // year, two digits
                    string.append(KGlobal::locale()->calendar()->yearString(DateTime.date(), KCalendarSystem::ShortFormat));
                break;
                case 'Z': // timezone city
                    string.append(prettyTimezone());
                break;
                default:
                    string.append(format.at(index));
                break;
            }
        }
        else
        {
            string.append(format.at(index));
        }
    }

    if (format.at(length - 1) != '%')
    {
        string.append(format.at(length));
    }

    return string;
}

QList<QAction*> AdjustableClock::contextualActions()
{
    return m_actions;
}

#include "plasma-adjustableclock.moc"
